import json
import logging
from aiohttp.web import Request, WebSocketResponse
from aiohttp import WSMsgType
from pychromecast import get_chromecasts
import asyncio
import datetime

logger = logging.getLogger(__name__)

def default(o):
    if isinstance(o, (datetime.date, datetime.datetime)):
        return o.isoformat()


class ChromecastListener:
    def __init__(self, ws: WebSocketResponse):
        self.ws = ws
        self._status = {'cast': None, 'media': None}

    def send_json(self):
        loop = asyncio.new_event_loop()
        task = loop.create_task(self.ws.send_str(json.dumps(self._status, default = default)))
        loop.run_until_complete(task)

    def new_cast_status(self, status):
        self._status['cast'] = status._asdict()
        self.send_json()

    def new_media_status(self, status):
        self._status['media'] = status.__dict__
        self.send_json()

async def chromecast_proxy_service(request: Request):

    ws = WebSocketResponse()

    await ws.prepare(request)

    friendly_name = request.match_info.get('friendly_name')

    logger.info("Client connected to chromecast proxy to %s", friendly_name)

    chromecast = next(cc for cc in get_chromecasts() if cc.device.friendly_name == friendly_name)

    chromecast.start()

    cc_listener = ChromecastListener(ws)
    chromecast.register_status_listener(cc_listener)
    chromecast.media_controller.register_status_listener(cc_listener)

    async for msg in ws:
        if msg.type == WSMsgType.TEXT:
            if msg.data == 'close':
                await ws.close()

        elif msg.type == WSMsgType.ERROR:
            logger.error("ws connection closed with exception %s", ws.exception())

    logger.info("Chromecast proxy websocket connection closed")

    return ws



import asyncio
import json
import logging
import re
import subprocess
from datetime import date, timedelta
from functools import wraps
from io import StringIO
import pychromecast


from aiohttp import ClientSession, web, BasicAuth
from lxml import etree

from .controller_base import ControllerBase
from .controller_cache import ControllerCache
from .tools import parse_temp, tree_search
from .transmission import Transmission, TransmissionError

logger = logging.getLogger(__name__)

ping_pattern = re.compile(r'64 bytes from ([0-9\.]{7,15}): icmp_.eq=([0-9]{1,2}) ttl=([0-9]{1,4}) time=([0-9\.]{1,7}) ms')

parser = etree.HTMLParser()

def route(path, method = 'GET'):
    def decor(func):
        func._route_data = (path, method)
        return func
    return decor

class CalledProcessorError(Exception):

    def __init__(self, code, stdout, stderr):
        self.code = code
        self.stdout = stdout
        self.stderr = stderr
        super().__init__("CalledProcessorError code %s" % code)

async def check_output(cmd):
    proc = await asyncio.create_subprocess_shell(cmd, stdout = asyncio.subprocess.PIPE, stderr = asyncio.subprocess.PIPE)
    stdout, stderr = await proc.communicate()
    if proc.returncode != 0:
        raise CalledProcessorError(proc.returncode, stdout, stderr)
    return stdout


async def get_json(url):
    async with ClientSession() as session:
        async with session.get(url) as response:
            return await response.json(content_type = None)

async def get_xml(url):
    async with ClientSession() as session:
        async with session.get(url) as response:
            text = await response.text()
            return etree.parse(StringIO(text), parser)

cache = ControllerCache()

class Api(ControllerBase):

    def __init__(self, app: web.Application):
        for name in dir(self):
            func = getattr(self, name)
            if not hasattr(func, '_route_data'):
                continue
            (path, method) = func._route_data

            app.router.add_route(method, '/api' + path, func)

    @route('/server-status', 'POST')
    async def server_status(self, request: web.Request):

        config = await request.json()

        ip = config['ip']
        status_server_port = config['statusServerPort']

        cnt = 4

        logger.debug("Pinging %s ...", ip)
        try:
            ping_raw = await check_output('ping %s -c %d -i 0.2' % (ip, cnt))
        except CalledProcessorError as e:
            logger.debug("Ping failed %s", e)
            return web.json_response({"ping": None})

        time_sum = 0
        for match in re.finditer(ping_pattern, ping_raw.decode()):
            time_sum += float(match.group(4))

        try:
            server_details = await get_json('http://%s:%s/' % (ip, status_server_port))
        except Exception as e:
            logger.info("Cannot reach %s:%s because %s", ip, status_server_port, e)
            server_details = {}

        return web.json_response({
            'ping': round(time_sum / cnt, 1),
            **server_details,
        })

    @route('/idokep/current/{city}')
    @cache.cached_json_controller(timedelta(minutes = 10))
    async def idokep_current(self, request: web.Request):

        city = request.match_info.get('city')

        base_url = 'https://www.idokep.hu'

        logger.debug("Fetch data from %s", base_url)

        tree = await get_xml('%s/idojaras/%s' % (base_url, city))

        current_container = tree_search('.fooldal-elorejelzes .jelenlegi', tree)

        res = {
            'img': base_url + tree_search('.icon > img', current_container).attrib['src'],
            'value': parse_temp(tree_search('.homerseklet', current_container).text),
        }

        return res


    @route('/idokep/days/{city}')
    @cache.cached_json_controller(timedelta(hours = 3))
    async def idokep_days(self, request):

        base_url = 'https://www.idokep.hu'

        logger.debug("Fetch data from %s", base_url)
        tree = await get_xml('%s/elorejelzes/%s' % (base_url, request.match_info.get('city')))

        day_columns = tree_search('.huszonegy .oszlop', tree, False)

        res = []

        col_date = None

        for day_column in day_columns:
            day_cell = tree_search('.nap-box > div', day_column)
            if day_cell is None:
                continue

            precipitation_val = 0
            precipitation_prob = 0

            day = int(day_cell.text)

            if col_date is None:
                col_date = date.today()
                while col_date.day != day:
                    col_date = col_date + timedelta(days = 1)

            csapadek_tree = tree_search('.csapadek', day_column)
            if csapadek_tree is not None:
                prec_val_text = tree_search('.buborek-lent .buborek-sor .buborek-text .csapadek-text', csapadek_tree).text
                prec_prob_text = tree_search('.buborek-lent .buborek-sor .buborek-text .valoszinuseg', csapadek_tree).text
                precipitation_matches = re.search(r'Kb\. ([\d]{1,}) mm', prec_val_text)
                precipitation_val = precipitation_matches.group(1) if precipitation_matches else 0
                precipitation_prob = re.search(r'.+ ([\d]{1,})%', prec_prob_text).group(1)

            day_data = {
                # 'img': base_url + tree_search('.icon > svg > image', day_column).attrib['xlink:href'],
                'img': base_url + tree_search('.icon > img', day_column).attrib['src'],
                'day': day,
                'date': str(col_date),
                'max': int(tree_search('[class^="max-homerseklet-"]', day_column).text.strip()),
                'min': int(tree_search('[class^="min-homerseklet-"]', day_column).text.strip()),
                'precipitation': {
                    'value': int(precipitation_val),
                    'probability': int(precipitation_prob),
                },
            }
            res.append(day_data)

            col_date = col_date + timedelta(days = 1)

        return res

    @route('/idokep/hours/{city}')
    @cache.cached_json_controller(timedelta(minutes = 30))
    async def idokep_hours(self, request):

        base_url = 'https://www.idokep.hu'

        logger.debug("Fetch data from %s", base_url)
        tree = await get_xml('%s/elorejelzes/%s' % (base_url, request.match_info.get('city')))

        day_columns = tree_search('.tizenket .oszlop', tree, False)

        res = []

        for day_column in day_columns:
            hour_cell = tree_search('.ora', day_column)
            if hour_cell is None:
                continue

            prec_val_tree = tree_search('.csapadek-text', day_column)
            prec_prob_tree = tree_search('.valoszinuseg', day_column)
            precipitation_val = 0
            precipitation_prob = 0

            if prec_val_tree is not None:
                precipitation_val = float(re.search(r'Kb\. ([\.\d]{1,}) mm', prec_val_tree.text).group(1))

            if prec_prob_tree is not None:
                precipitation_prob = float(re.search(r'.+ ([\.\d]{1,})%', prec_prob_tree.text).group(1))

            wind_element = tree_search('.szel-icon > img', day_column)

            hour_data = {
                'hour': int(hour_cell.text[:-1]),
                'img': base_url + tree_search('.icon > img', day_column).attrib['src'],
                'temp': int(tree_search('.hoerzet', day_column).text),
                'precipitation': {
                    'value': precipitation_val,
                    'probability': precipitation_prob,
                },
                'wind': {
                    'img': base_url + wind_element.attrib['src'],
                    'angle': int(re.search(r'rotate\(([\d]{1,3})deg\)', wind_element.attrib['style']).group(1)),
                },
            }

            res.append(hour_data)

        return res

    @route('/chromeasts')
    @cache.cached_json_controller(timedelta(minutes = 5))
    async def chromecast_list(self, request):
        chromecasts = pychromecast.get_chromecasts()

        return [cc.device.friendly_name for cc in chromecasts]

    @route('/transmission', 'POST')
    async def transmission(self, request):

        config = await request.json()

        client = Transmission(config['url'], config.get('username'), config.get('password'))

        try:
            return web.json_response({
                'data': {
                    'session-stats': await client.session_stats(),
                    'torrents': await client.torrent_list(),
                },
                'status': 'success'
            })
        except TransmissionError as e:
            return web.json_response({
                'data': str(e),
                'status': 'error'
            })



from pickle import dumps, loads
from pymemcache.client.base import Client

MemcacheClient = Client

def memcache_serializer(key, value):
    return dumps(value), 1

def memcache_deserializer(key, value, flags):
    return loads(value)

def create_memcache_client(host, port) -> MemcacheClient:
    return Client((host, port), serializer = memcache_serializer, deserializer = memcache_deserializer)



import os
import logging
from pathlib import Path

from enum import Enum

from configpp.soil import Config, Transport, Location

class Mode(Enum):

    DEVELOPMENT = 'development'
    PRODUCTION = 'production'

logger = logging.getLogger(__name__)


default_dashboard_data = {
    'dashboards': {},
    'widgets': {},
}

class AppConfig:

    dashboard_config_loader: Config

    def load(self, path_like: str):

        path = Path(path_like).expanduser().absolute()

        if not path.parent.is_dir():
            path.parent.mkdir(parents = True)

        self._location = Location(path.parent)

        self.dashboard_config_loader = Config(path.name, transport = Transport([self._location]))

        return self.dashboard_config_loader.load()

    def save(self):
        self.dashboard_config_loader.dump(self._location)

    def init_logger(self, debug: bool):
        logging.basicConfig(
            level = logging.DEBUG if debug else logging.INFO,
            format = '%(asctime)s - %(levelname)s - %(filename)s:%(lineno)s: %(message)s',
        )



import asyncio
import logging
import os
from time import sleep, time

import pkg_resources
from aiohttp import web
from jinja2 import Template

from .api import Api
from .config_sync_service import ConfigSyncService
from .javascript_libraries import javascript_libraries
from .config import AppConfig, Mode, default_dashboard_data
from .chromecast_proxy_service import chromecast_proxy_service

logger = logging.getLogger(__name__)

class App:

    def __init__(self, database_file: str, debug: bool):

        self.config = AppConfig()
        self.config.init_logger(debug)
        self.config.load(database_file)
        self.debug = debug

        self.aio_app = web.Application()

        current_directory = os.path.dirname(__file__)

        with open(os.path.join(current_directory, 'templates/index.html')) as f:
            self._index_template = Template(f.read())


        self.api = Api(self.aio_app)
        self.config_syncer = ConfigSyncService(self.config)

        self.aio_app.router.add_get('/', self.index)
        self.aio_app.router.add_static('/static', os.path.join(current_directory, 'static'))
        self.aio_app.router.add_route('GET', '/config-sync', self.config_syncer.controller)
        self.aio_app.router.add_route('GET', '/chromecast-proxy/{friendly_name}', chromecast_proxy_service)

    async def index(self, request):

        mode = Mode.DEVELOPMENT if self.debug else Mode.PRODUCTION

        version = str(time()) if self.debug else pkg_resources.get_distribution("claude").version

        resp_text = self._index_template.render(
            version = version,
            dev_mode = mode == Mode.DEVELOPMENT,
            initial_data = self.config.dashboard_config_loader.data or default_dashboard_data,
            javascript_libraries = javascript_libraries[mode],
        )

        return web.Response(body = resp_text, content_type = 'text/html')



import logging
from dataclasses import dataclass
from datetime import datetime, timedelta
from functools import wraps
from typing import Any, Dict

from aiohttp import web

from .controller_base import ControllerBase

logger = logging.getLogger(__name__)

class ControllerCache:

    @dataclass
    class Item:
        data: Any
        expired: datetime

    def __init__(self):
        self._data: Dict[str, self.Item] = {}

    def set(self, key: str, data, expiry: timedelta):
        self._data[key] = self.Item(data, datetime.now() + expiry)

    def get(self, key: str):
        item = self._data.get(key)
        if item is None:
            return None
        if item.expired < datetime.now():
            return None
        return item.data

    def cached_json_controller(self, expiry: timedelta):
        def wrapper(func):
            @wraps(func)
            async def decor(controller: ControllerBase, request: web.Request):
                data = self.get(request.path)
                if data and not request.query.get('force-refetch') and expiry.total_seconds() > 0:
                    logger.debug("Get '%s' data from cache", request.path)
                    return web.json_response(data)

                data = await func(controller, request)

                self.set(request.path, data, expiry)

                return web.json_response(data)
            return decor
        return wrapper




import json
from aiohttp import ClientSession, BasicAuth, ClientError

SESSION_ID_HEADER_NAME = 'X-Transmission-Session-Id'

class TransmissionError(Exception):
    pass

class Transmission:

    def __init__(self, url, username = None, password = None):
        self._url = url
        self._username = username
        self._password = password
        self._session_id = ''


    async def torrent_list(self):
        data = await self._call('torrent-get', {
            "fields": [ "id", "name", "totalSize", "percentDone", "rateDownload", "sizeWhenDone", "eta", "rateUpload", "peersConnected" ],
        })
        return data['arguments']['torrents']

    async def session_stats(self):
        data = await self._call('session-stats')
        return data['arguments']

    async def _call(self, method: str, arguments: dict = {}):

        return await self._request({
            "arguments": arguments,
            "method": method,
        })

    async def _request(self, data: dict):

        headers = {
            SESSION_ID_HEADER_NAME: self._session_id,
        }

        auth = BasicAuth(self._username, self._password) if (self._username and self._password) else None

        try:

            async with ClientSession() as session:
                async with session.post(self._url, data = json.dumps(data), headers = headers, auth = auth) as response:

                    if response.status == 409 and self._session_id == '':
                        self._session_id = response.headers[SESSION_ID_HEADER_NAME]
                        return await self._request(data)

                    if response.status != 200:
                        raise TransmissionError(await response.text())

                    return await response.json()

        except ClientError as e:
            raise TransmissionError(str(e))



from .config import Mode

javascript_libraries = {
    Mode.DEVELOPMENT: [{
            'url': 'https://cdnjs.cloudflare.com/ajax/libs/react/16.8.6/umd/react.development.js',
            'hash': 'sha256-9H05eCUa5x4G28dytYBwoAXkR3XgPE3sCogyXhrINwo=',
        },{
            'url': 'https://cdnjs.cloudflare.com/ajax/libs/react-dom/16.8.4/umd/react-dom.development.js',
            'hash': 'sha256-zaBIa6kuwEspzOIU5TeB5bATnxmvJFmht9smi5jLH/Q=',
        },{
            'url': 'https://unpkg.com/@material-ui/core@4.3.1/umd/material-ui.development.js',
            'hash': 'sha256-upouvKCn4yjDmMAMdOYw4RwWGG9xNm7KifNX81sp9Ag=',
        },{
            'url': 'https://cdn.jsdelivr.net/npm/@fortawesome/fontawesome-svg-core@1.2.21/index.js',
            'hash': 'sha256-v0YHQ2mwbxSbi6TzwAlY3vjJFG3FN9HRJGmXE/hCecs=',
        }, {
            'url': 'https://cdn.jsdelivr.net/npm/@fortawesome/free-solid-svg-icons@5.9.0/index.js',
            'hash': 'sha256-oSkdBl+Zjb67LV7yVUSCgWSMq3bfHzJ9Jz5l+Te4Y6s=',
        }, {
            'url': 'https://cdn.jsdelivr.net/npm/@fortawesome/free-brands-svg-icons@5.9.0/index.js',
            'hash': 'sha256-gMLZ7GnEPLSNwoaO0YtgOrhQkNIzA7rcvWsAtz3uolY=',
        }, {
            'url': 'https://cdnjs.cloudflare.com/ajax/libs/moment.js/2.24.0/moment-with-locales.js',
            'hash': 'sha256-dgFbqbQVzjkZPQxWd8PBtzGiRBhChc4I2wO/q/s+Xeo=',
        }, {
            'url': 'https://cdn.jsdelivr.net/npm/react-draggable@3.3.0/dist/react-draggable.js',
            'hash': 'sha256-dx6xPZVJdlAEbit+OnKSKb1JOwcJ7vyMdwatshRtm/U=',
        }, {
            'url': 'https://unpkg.com//prop-types@15.7.2/prop-types.min.js',
            'hash': 'sha256-TIg1BRfugqpPM2jmfvGkU8pmNtz6ZEm049b6pch3Bm4=',
        }, {
            'url': 'https://unpkg.com/recharts@1.7.1/umd/Recharts.min.js',
            'hash': 'sha256-1UBekyQ4ydBFe1qngtBJZUc0NhsSLbrl85xJxaNsl1o=',
        }, {
            'url': 'https://unpkg.com/react-redux@7.1.1/dist/react-redux.js',
            'hash': 'sha256-qhF8B/fGZJB8suWxMxiopLlx9uz8ZnF/nox4zirz7MM=',
        }],
    Mode.PRODUCTION: [{
            'url': 'https://cdnjs.cloudflare.com/ajax/libs/react/16.8.6/umd/react.production.min.js',
            'hash': 'sha256-3vo65ZXn5pfsCfGM5H55X+SmwJHBlyNHPwRmWAPgJnM=',
        },{
            'url': 'https://cdnjs.cloudflare.com/ajax/libs/react-dom/16.8.6/umd/react-dom.production.min.js',
            'hash': 'sha256-qVsF1ftL3vUq8RFOLwPnKimXOLo72xguDliIxeffHRc=',
        },{
            'url': 'https://unpkg.com/@material-ui/core@4.3.1/umd/material-ui.production.min.js',
            'hash': 'sha256-Ulx0cEzoGy7n7MfY5i4p1eMoYYMvoiS4aE2WpSQ2o1M=',
        },{
            'url': 'https://cdn.jsdelivr.net/npm/@fortawesome/fontawesome-svg-core@1.2.21/index.min.js',
            'hash': 'sha256-rP06f4bKRPwHs7l65xPApnEtLhaPvduxlgoBKgIzeSY=',
        }, {
            'url': 'https://cdn.jsdelivr.net/npm/@fortawesome/free-solid-svg-icons@5.9.0/index.min.js',
            'hash': 'sha256-5sSsNI4hqSRL1Sc1+a3neqzFf3RxwDqyRaWb3vB1evE=',
        }, {
            'url': 'https://cdn.jsdelivr.net/npm/@fortawesome/free-brands-svg-icons@5.9.0/index.min.js',
            'hash': 'sha256-KR+l1PmbxB3Cipzs6QhT8E0KxMaKsRzCl1Pf9qSgO8c=',
        }, {
            'url': 'https://cdnjs.cloudflare.com/ajax/libs/moment.js/2.24.0/moment-with-locales.min.js',
            'hash': 'sha256-AdQN98MVZs44Eq2yTwtoKufhnU+uZ7v2kXnD5vqzZVo=',
        }, {
            'url': 'https://cdn.jsdelivr.net/npm/react-draggable@3.3.2/dist/react-draggable.min.js',
            'hash': 'sha256-Bz00T+A2ZjG2P8ULO5VodkZd/JKF0sAHWFxz+uBRkxE=',
        }, {
            'url': 'https://unpkg.com/prop-types@15.7.2/prop-types.min.js',
            'hash': 'sha256-TIg1BRfugqpPM2jmfvGkU8pmNtz6ZEm049b6pch3Bm4=',
        }, {
            'url': 'https://unpkg.com/recharts@1.7.1/umd/Recharts.min.js',
            'hash': 'sha256-1UBekyQ4ydBFe1qngtBJZUc0NhsSLbrl85xJxaNsl1o=',
        }, {
            'url': 'https://unpkg.com/react-redux@7.1.1/dist/react-redux.min.js',
            'hash': 'sha256-N8z+GQ8ps7fpXZpwEPF6KnnjpQoH/hB0zC1DXJLTM9w=',
        }],
}



import json
import logging
from aiohttp.web import Request, WebSocketResponse
from aiohttp import WSMsgType
from time import time

from .config import default_dashboard_data, AppConfig
from .tools import dict_merge

logger = logging.getLogger(__name__)

clients = []

def find_removed_node(node: dict, data: dict):
    for key, value in node.items():
        if value is None:
            del data[key]
        else:
            find_removed_node(value, data[key])

class ConfigSyncService:

    def __init__(self, config: AppConfig):
        self._config = config

    async def controller(self, request: Request):

        ws = WebSocketResponse()

        await ws.prepare(request)

        clients.append(ws)

        logger.info("New client connected to the sync service")

        async for msg in ws:
            if msg.type == WSMsgType.TEXT:
                if msg.data == 'close':
                    await ws.close()
                else:
                    message = json.loads(msg.data)

                    action = message['action']
                    diff = message['diff']

                    if not self._config.dashboard_config_loader.is_loaded:
                        self._config.dashboard_config_loader.data = default_dashboard_data

                    dict_merge(diff['added'], self._config.dashboard_config_loader.data)
                    dict_merge(diff['updated'], self._config.dashboard_config_loader.data)

                    find_removed_node(diff['deleted'], self._config.dashboard_config_loader.data)

                    # TODO: delay with some sec (and add if new trigger)
                    self._config.save()

                    dispatch_clients = list(filter(lambda c: c != ws, clients))

                    if dispatch_clients:

                        action['time'] = time()
                        action_msg = json.dumps(action)

                        logger.debug("Action dispatch to %s clients(s)", len(dispatch_clients))

                        for client in dispatch_clients:
                            await client.send_str(action_msg)

            elif msg.type == WSMsgType.ERROR:
                logger.error("ws connection closed with exception %s", ws.exception())

        logger.info("websocket connection closed")

        clients.remove(ws)

        return ws



import logging
import re

from lxml.cssselect import CSSSelector

TEMP_PATTERN = re.compile(r'([\d+\-]+)')

logger = logging.getLogger(__name__)

def dict_merge(source, destination):
    """
    run me with nosetests --with-doctest file.py

    >>> a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
    >>> b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
    >>> dict_merge(b, a) == { 'first' : { 'all_rows' : { 'pass' : 'dog', 'fail' : 'cat', 'number' : '5' } } }
    True
    """
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            dict_merge(value, node)
        else:
            destination[key] = value

    return destination


def parse_temp(temp: str) -> int:
    """
    >>> parse_temp("-1")
    -1
    >>> parse_temp("+42°F")
    42
    >>> parse_temp(" -42 ° C ")
    -42
    """

    logger.debug("parse temp: '%s'", temp)

    m = TEMP_PATTERN.search(temp)

    return int(m.group(1)) if m else None

def tree_search(selector, tree, return_first = True):
    sel = CSSSelector(selector)
    res = sel(tree)
    if not len(res):
        return None
    return res[0] if return_first else res



from setuptools import setup, find_packages

setup(
    name = "claude",
    description = "Pluginable dashboard app",
    long_description = "Pluginable dashboard app",
    author = "Lajos Santa",
    author_email = "santa.lajos@gmail.com",
    url = "https://github.com/voidpp/claude",
    license = "MIT",
    use_scm_version = True,
    setup_requires = ["setuptools_scm"],
    install_requires = [
        "aiohttp~=3.5",
        "Jinja2~=2.10",
        "configpp~=0.3",
        "lxml~=4.4",
        "pymemcache~=2.2",
        "cssselect~=1.1",
        "cachetools~=3.1",
        "pychromecast~=3.2",
    ],
    include_package_data = True,
    packages = find_packages(),
    scripts = [
        "bin/start-claude-server"
    ],
    extras_require = {},
    classifiers = [],
)
